import { Monad } from "./fantasy";
import { OptionalAsync } from "./optional-async";
export declare type Nil = undefined | null;
/**
 * A container object which may or may not contain a non-null value.
 * If a value is present, isEmpty() will return true and get() will return the value.
 *
 * Some operations you can do:
 * ```
 * const result = Optional.of(1)
 *   .map(i => i + 1)                   // => Some(2)
 *   .map(i => i * i)                   // => Some(4)
 *   .flatMap(i => Optional.of(i - 1))  // => Some(3)
 *   .pMap(i => Promise.resolve(i));    // => Promise(Some(3))
 *
 * const result2 = await result;        // => Some(3)
 * result2.isEmpty()                    // => false
 * result2.contains(3)                  // => true
 * result2.get()                        // => 3
 *
 * const result3 = result2              // => Some(3)
 *   .flatMap(i => Optional.empty())    // => None
 * result3.get()                        // => throws Error()
 * result3.getOrElse(4)                 // => 4
 * ```
 */
export declare abstract class Optional<A> implements Monad<A> {
    static isNil<T>(t: T | null | undefined): t is Nil;
    static of<B>(v: B | null | undefined): Optional<B>;
    static empty<B>(): Optional<B>;
    static all<B>(optxs: Optional<B>[]): Optional<B[]>;
    /**
     * returns an array that keeps only the non empty values of an array of Optionals
     * ```
     * const arr = [1, 2, undefined, 4, null, 6].map(Optional.of);
     * Optional.flatten(arr) // => [1, 2, 4, 6]
     * ```
     * @param {Optional<B>[]} arr
     */
    static flatten<B>(arr: Optional<B>[]): B[];
    static contains<A>(value: A): (opt: Optional<A>) => boolean;
    static contains<A>(value: A, opt: Optional<A>): boolean;
    static isEmpty<A>(opt: Optional<A>): boolean;
    abstract ap<B>(b: Optional<(a: A) => B>): Optional<B>;
    /**
     * applies the function to the inner value if it exists.
     *
     * ```
     * const square = (v) => v * v;
     * const squareOf11 = Optional.of(11).map(square)
     * squareOf11.contains(121) // => true
     * ```
     *
     * @param {(A) => B} fn
     * @returns {Optional<B>}
     */
    abstract map<B>(fn: (a: A) => B | null | undefined): Optional<B>;
    /**
     * applies the async function to the inner value if it exists. Swaps what
     * would have been an inner Promise to the outside.
     * @param {(A) => Promise<B>} fn
     * @returns {Promise<Optional<B>>}
     * @deprecated
     */
    abstract pMap<B>(fn: (a: A) => Promise<B>): Promise<Optional<B>>;
    /**
     * applies the function to the inner value if it exists.
     * similar to `map` except the function should return an Optional and the final
     * result will be empty if the function returns an empty Optional.
     *
     * ```
     * Optional.of("hello")
     *   .flatMap(str => Optional.of(str + " world")) // => Some("hello world")
     *   .flatMap(str => Optional.of(null))           // => None
     * ```
     * @param {(A) => Optional<B>} fn
     * @returns {Optional<B>}
     */
    abstract flatMap<B>(fn: (a: A) => Optional<B>): Optional<B>;
    chain<B>(fn: (a: A) => Optional<B>): Optional<B>;
    /**
     * returns the inner value
     *
     * ```
     * Optional.of(undefined).get()     // => throws Error
     * Optional.of(null).get()          // => throws Error
     * Optional.of("hello world").get() // => "hello world"
     * ```
     *
     * @throws {Error} when trying to get a value of an empty Optional
     */
    abstract get(): A;
    /**
     * returns the inner value if it exists, otherwise returns the passed in value
     *
     * ```
     * Optional.empty().getOrElse("default")            // => "default"
     * Optional.of("hello world").getOrElse("default")  // => "hello world"
     * ```
     *
     * @param d the default value to return if the Optional is empty
     */
    abstract getOrElse(d: A): A;
    /**
     * returns the inner value if it exists, otherwise throws with given error
     * ```
     * const error = new Error("my descriptive error")
     * Optional.empty().getOrThrow(error)             // => ERROR: Error: my descriptive error
     * Optional.of("hellow world").getOrThrow(error)  // => "hello world"
     * ```
     * @param {Error} error - error object to throw
     */
    abstract getOrThrow(error: Error): A;
    /**
     * returns the inner value if it exists, otherwise returns null
     * ```
     * Optional.empty().getOrNull()     // => null
     * Optional.of("HELLO").getOrNull() // => "HELLO"
     */
    abstract getOrNull(): A | null;
    /**
     * returns the inner value if it exists, otherwise returns undefined
     * ```
     * Optional.empty().getOrNull()     // => undefined
     * Optional.of("HELLO").getOrNull() // => "HELLO"
     */
    abstract getOrUndefined(): A | undefined;
    /**
     * returns whether or not the Optional's inner value is equal to what is
     * passed in. Uses `===`
     *
     * ```
     * Optional.of("hello world").contains("hello world")   // => true
     * Optional.empty().contains("hello world")             // => false
     * ```
     * @param v
     */
    abstract contains(v: A): boolean;
    /**
     * returns true if this Optional contains `undefined` or `null`, false otherwise.
     *
     * ```
     * Optional.of("hello world").isEmpty() // => false
     * Optional.of(0).isEmpty()             // => false
     * Optional.of({}).isEmpty()            // => false
     * Optional.of([]).isEmpty()            // => false
     * Optional.empty().isEmpty()           // => true
     * Optional.of(null).isEmpty()          // => true
     * Optional.of(undefined).isEmpty()     // => true
     *
     * ```
     */
    abstract isEmpty(): boolean;
    /**
     * Returns true if this Optional is not empty and the function passed in returns true. Otherwise
     * returns false
     * @param {(A) => boolean} fn - existance function
     */
    abstract exists(fn: (v: A) => boolean): boolean;
    /**
     * Converts this Optional into an OptionalAsync to make it easier to work with async functions.
     */
    abstract async(): OptionalAsync<A>;
    /**
     * Applies an async function to the inner value and converts the Optional into an OptionalAsync. This
     * is an alias for calling {@link async} and then {@link OptionalAsync.mapAsync}.
     * @param {(v: A) => Promise<B>} fn
     */
    mapAsync<B>(fn: (v: A) => Promise<B>): OptionalAsync<B>;
    abstract case<B>(predicates: {
        some: (value: A) => Optional<B>;
        none: () => Optional<B>;
    }): Optional<B>;
}
export declare class Some<A> extends Optional<A> {
    private readonly a;
    constructor(v: A);
    ap<B>(b: Optional<(a: A) => B>): Optional<B>;
    contains(v: A): boolean;
    exists(fn: (v: A) => boolean): boolean;
    flatMap<B>(fn: (a: A) => Optional<B>): Optional<B>;
    get(): A;
    getOrElse(): A;
    getOrThrow(): A;
    getOrNull(): A | null;
    getOrUndefined(): A | undefined;
    isEmpty(): boolean;
    map<B>(fn: (a: A) => B | null | undefined): Optional<B>;
    pMap<B>(fn: (a: A) => Promise<B>): Promise<Optional<B>>;
    async(): OptionalAsync<A>;
    case<B>(predicates: {
        some: (value: A) => Optional<B>;
        none: () => Optional<B>;
    }): Optional<B>;
}
export declare class None<A> extends Optional<A> {
    constructor();
    ap<B>(): Optional<B>;
    contains(): boolean;
    exists(): boolean;
    flatMap<B>(): Optional<B>;
    get(): A;
    getOrElse(d: A): A;
    getOrThrow(error: Error): A;
    getOrNull(): A | null;
    getOrUndefined(): A | undefined;
    isEmpty(): boolean;
    map<B>(): Optional<B>;
    pMap<B>(): Promise<Optional<B>>;
    async(): OptionalAsync<A>;
    case<B>(predicates: {
        some: (value: A) => Optional<B>;
        none: () => Optional<B>;
    }): Optional<B>;
}
/** Alias for Optional */
export declare const Opt: typeof Optional;
