import { Monad } from "./fantasy";
import { Optional } from "./optional";
import { TryAsync } from "./try-async";
/**
 * Try is a monadic container type which represents a computation that may either result in an exception,
 * or return a successfully computed value. Instances of Try, are either an instance of Success or Failure.
 *
 * An important property of Try is its ability to pipeline, or chain, operations, catching exceptions along the way.
 * The {@link Try.flatMap} and {@link Try.map} combinators each essentially pass off either their successfully completed
 * value, wrapped in a Success type for it to be further operated upon by the next combinator in the chain, or the
 * exception wrapped in a Failure type usually to be simply passed on down the chain. Combinators such as
 * {@link Try.recover} are designed to provide some type of default behavior in the case of failure.
 *
 * lets say you have a function that might return an error:
 * ```
 * function foobar(): number {
 *   const n = Math.random();
 *   return n > 0.5 ? n : throw new Error("you were unlucky");
 * }
 * ```
 *
 * case 1: foobar() returns 0.7
 * ```
 * const result = Try.of(foobar)        // => Success(0.7)
 *   .map((n) => n + 1)                 // => Success(1.7)
 *   .flatMap((n) => Try.success(n -2)) // => Success(-1.7)
 *   .recover((e) => 0.0)               // => Success(-1.7)
 *   .get();                            // => -1.7
 *
 * assert(result, -1.7)
 * ```
 *
 * case 2: foobar() throws an error
 * ```
 * const result = Try.of(foobar)        // => Failure("you were unlucky")
 *   .map((n) => n + 1)                 // => Failure("you were unlucky")
 *   .flatMap((n) => Try.success(n -2)) // => Failure("you were unlucky")
 *   .recover((e) => 0.0)               // => Success(0.0)
 *   .get();                            // => 0.0
 * ```
 */
export declare abstract class Try<A> implements Monad<A> {
    static isSuccess<B>(t: Try<B>): t is Success<B>;
    static isFailure<B>(t: Try<B>): t is Failure<B>;
    static of<B>(fn: () => B): Try<B>;
    static pOf<B>(fn: () => Promise<B>): Promise<Try<B>>;
    /**
     * returns an array that keeps only the successful values of an array of Try
     * ```
     * const arr = [() => 1, () => throw new Error()].map(Try.of);
     * Try.flatten(arr) // => [1]
     * ```
     * @param {Try<B>[]} arr
     */
    static flatten<B>(arr: Try<B>[]): B[];
    /**
     * Takes an array of Try<B> and returns a single Try containing all the inner values as B[]
     * if they are all successful.
     *
     * @example Try.all([Try.success(1), Try.success(2), Try.success(3)])     => Try.success([1, 2, 3])
     * @example Try.all([Try.success(1), Try.failure(error), Try.success(3)]  => Try.failure(error)
     * @param {Try<B>[]>} arr - array of Try
     */
    static all<B>(arr: Try<B>[]): Try<B[]>;
    /**
     * builds an instance of Failure directly
     * ```
     * Try.failure(new Error("error message"));
     * ```
     * @param {Error} error
     * @return {Try<B>}
     */
    static failure<B>(error: Error): Try<B>;
    /**
     * builds an instance of Success directly
     * ```
     * Try.success({foo: "bar"});
     * ```
     * @param {B} b
     * @return {Try<B>}
     */
    static success<B>(b: B): Try<B>;
    /**
     * utility for Try.success<void>(undefined)
     * ```
     * Try.void();
     * ```
     * @return {Try<void>}
     */
    static void(): Try<void>;
    abstract readonly isSuccess: () => this is Success<A>;
    abstract readonly isFailure: () => this is Failure<A>;
    abstract readonly get: () => A;
    abstract readonly getOrElse: (defaultValue: A) => A;
    abstract readonly toOptional: () => Optional<A>;
    abstract ap<B>(b: Try<(a: A) => B>): Try<B>;
    /**
     * applies the function to the success value
     * ```
     * const fn = (str) => str + " world";
     * Try.success("hello").map(fn);                // Success("hello world")
     * Try.failure<string>(new Error()).map(fn);    // Failure()
     * ```
     * @param {(a: A) => B} fn - function to apply to the success value
     * @return {Try<B>}
     */
    abstract readonly map: <B>(fn: (a: A) => B) => Try<B>;
    /**
     * applies the function that returns a `Try` to the success value and flattens over
     * the result so you don't get a `Try` within a `Try`.
     *
     *```
     * const fn = (str) => Try.success(str + "world");
     * Try.success("hello").flatMap(fn);                    // Success("hello world");
     * Try.failure<string>(new Error()).flatMap(fn);        // Failure()
     *
     * @param {(a: A) => Try<B>} fn - function to apply to the success value
     * @return {Try<B>}
     */
    abstract readonly flatMap: <B>(fn: (a: A) => Try<B>) => Try<B>;
    /** @see {@link flatMap} */
    chain<B>(fn: (a: A) => Try<B>): Try<B>;
    /**
     * simliar to map except the function being applied returns a {@link Promise}.
     * This is useful if you need to return something async inside the mapping function.
     *
     * ```
     * const fn = (str: string): Promise<string> => Promise.resolve(str + " world");
     * Try.success("hello").pMap(fn); // Promise<Sucess("hello world")>
     * ```
     *
     * If the mapping function throws, the result will automatically converted to a Failure.
     * If the mapping function's Promise rejects, the result will be converted toa Failure.
     * ```
     * Try.success("hello").pMap(async () => throw new Error());        // Promise<Failure>
     * Try.success("hello").pMap(() => Promise.reject(new Error()));    // Promise<Failure>
     * ```
     *
     * However, most of the time, you should just use {@link Try.async()} in order to deal with
     * asynchronous operations.
     * @deprecated
     */
    abstract readonly pMap: <B>(fn: (a: A) => Promise<B>) => Promise<Try<B>>;
    /**
     * similar to flatMap except the function being applied returns a Promise<Try>.
     * This is useful if the mapping function returns a Promise<Try>.
     * @deprecated
     */
    abstract readonly pFlatMap: <B>(fn: (a: A) => Promise<Try<B>>) => Promise<Try<B>>;
    /**
     * applies the function to an error. The function is only called if this is a Failure.
     * This is useful if you want to provide a default value on a failure based on the error.
     * ```
     * Try.failure(new Error()).recover(() => "Hello World!");  // Success("Hello World!")
     * ```
     */
    abstract readonly recover: (fn: (e: Error) => A) => Try<A>;
    /**
     * similar to {@link Try.recoverWith} except that the function returns a Try.
     * This is useful if you want remap the error into a different error or provide
     * a default value.
     * ```
     * const fn = (error: Error) => {
     *   return error.message === "bad bad bad" ? Try.success("okay") : Try.failure(new Error("not okay"));
     * }
     *
     * Try.failure(new Error("bad bad bad")).recoverWith(fn);   // Success("okay");
     * Try.failure(new Error("bad")).recoverWith(fn);           // Failure(new Error("not okay"));
     */
    abstract readonly recoverWith: (fn: (e: Error) => Try<A>) => Try<A>;
    /**
     * Maps the resolved value to `void` (undefined in this case). In other words, basically just throws away the
     * resolved value. If this is a {@link Failure}, then it will continue to pass along the failure.
     */
    void: () => Try<void>;
    abstract case<B>(predicates: {
        success: (value: A) => Try<B>;
        failure: (error: Error) => Try<B>;
    }): Try<B>;
    /**
     * Converts this Try<A> to a TryAsync<A>.
     */
    abstract readonly async: () => TryAsync<A>;
    /**
     * Applies an async function `(A) => Promise<B>` to the inner value and converts this Try<A> into a TryAsync<B>. This
     * is an alias for calling {@link async} and then {@link TryAsync.mapAsync}.
     * @param {(v: A) => Promise<B>} fn
     */
    mapAsync: <B>(fn: (v: A) => Promise<B>) => TryAsync<B>;
}
export declare class Failure<A> extends Try<A> {
    readonly error: Error;
    constructor(e: Error);
    readonly isSuccess: () => boolean;
    readonly isFailure: () => boolean;
    readonly get: () => A;
    readonly getOrElse: (defaultValue: A) => A;
    readonly toOptional: () => Optional<A>;
    readonly recover: (fn: (e: Error) => A) => Try<A>;
    readonly recoverWith: (fn: (e: Error) => Try<A>) => Try<A>;
    ap<B>(): Try<B>;
    readonly map: <B>() => Try<B>;
    readonly flatMap: <B>() => Try<B>;
    readonly pMap: <B>() => Promise<Try<B>>;
    readonly pFlatMap: <B>() => Promise<Try<B>>;
    case<B>(predicates: {
        success: (value: A) => Try<B>;
        failure: (error: Error) => Try<B>;
    }): Try<B>;
    readonly async: () => TryAsync<A>;
}
export declare class Success<A> extends Try<A> {
    readonly value: A;
    constructor(v: A);
    readonly isSuccess: () => boolean;
    readonly isFailure: () => boolean;
    readonly get: () => A;
    readonly getOrElse: () => A;
    readonly toOptional: () => Optional<A>;
    readonly recover: () => Try<A>;
    readonly recoverWith: () => Try<A>;
    ap<B>(b: Try<(a: A) => B>): Try<B>;
    readonly map: <B>(fn: (a: A) => B) => Try<B>;
    readonly flatMap: <B>(fn: (a: A) => Try<B>) => Try<B>;
    readonly pMap: <B>(fn: (a: A) => Promise<B>) => Promise<Try<B>>;
    readonly pFlatMap: <B>(fn: (a: A) => Promise<Try<B>>) => Promise<Try<B>>;
    case<B>(predicates: {
        success: (value: A) => Try<B>;
        failure: (error: Error) => Try<B>;
    }): Try<B>;
    readonly async: () => TryAsync<A>;
}
